# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2025 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea install` sub-command."""

import logging
from pathlib import Path
from typing import Any, Dict
from unittest import mock

import pytest
import yaml
from click import ClickException

from aea.configurations.base import DEFAULT_PROTOCOL_CONFIG_FILE
from aea.configurations.constants import DEFAULT_CONNECTION_CONFIG_FILE
from aea.test_tools.test_cases import AEATestCase, AEATestCaseEmpty

from tests.conftest import CUR_PATH


class TestInstall(AEATestCase):
    """Test that the command 'aea install' works as expected."""

    path_to_aea: Path = Path(CUR_PATH, "data", "dummy_aea")

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        result = self.run_cli_command("install", cwd=self._get_cwd())
        assert result.exit_code == 0

    def test_extra_dependencies(self, caplog: Any):
        """Assert that the exit code is equal to zero (i.e. success)."""
        with caplog.at_level(logging.DEBUG), mock.patch(
            "aea.cli.install.install_dependencies"
        ):
            result = self.run_cli_command(
                "-v",
                "DEBUG",
                "install",
                "-e",
                "open-aea-ledger-cosmos==1.0.0",
                cwd=self._get_cwd(),
            )

        assert result.exit_code == 0
        assert (
            "`['open-aea-ledger-cosmos<3.0.0,>=2.0.0']` will be overridden by ['open-aea-ledger-cosmos==1.0.0']"
            in caplog.text
        )


class TestInstallFromRequirementFile(AEATestCase):
    """Test that the command 'aea install --requirement REQ_FILE' works."""

    path_to_aea: Path = Path(CUR_PATH, "data", "dummy_aea")

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        super().setup_class()
        cls.result = cls.run_cli_command(
            "install", "-r", "requirements.txt", cwd=cls._get_cwd()
        )

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        assert self.result.exit_code == 0


class TestInstallFailsWhenDependencyDoesNotExist(AEATestCaseEmpty):
    """Test that the command 'aea install' fails when a dependency is not found."""

    capture_log = True

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        super().setup_class()
        result = cls.run_cli_command(
            "scaffold", "protocol", "-y", "my_protocol", cwd=cls._get_cwd()
        )
        assert result.exit_code == 0

        config_path = (
            Path(cls._get_cwd())
            / "protocols"
            / "my_protocol"
            / DEFAULT_PROTOCOL_CONFIG_FILE
        )
        with config_path.open() as fp:
            config = yaml.safe_load(fp)

        config.setdefault("dependencies", {}).update(
            {
                "this_is_a_test_dependency": {
                    "version": "==0.1.0",
                    "index": "https://test.pypi.org/simple",
                },
            }
        )

        with config_path.open(mode="w") as fp:
            yaml.safe_dump(config, fp)

    def test_error(self):
        """Assert an error occurs."""
        with pytest.raises(
            ClickException,
            match="An error occurred while installing.*this_is_a_test_dependency.*",
        ):
            self.run_cli_command("install", cwd=self._get_cwd())


class TestInstallWithRequirementFailsWhenFileIsBad(AEATestCase):
    """Test that the command 'aea install -r REQ_FILE' fails if the requirement file is not good."""

    path_to_aea: Path = Path(CUR_PATH, "data", "dummy_aea")

    def test_error(self):
        """Test that an error occurs."""
        with pytest.raises(
            ClickException,
            match="An error occurred while installing requirement file bad_requirements.txt. Stopping...",
        ):
            self.run_cli_command(
                "install", "-r", "bad_requirements.txt", cwd=self._get_cwd()
            )


class TestInstallFailsWhenDependencyHasUnsatisfiableSpecifier(AEATestCaseEmpty):
    """Test that the command 'aea install' fails when a dependency has an unsatisfiable version specifier."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        super().setup_class()
        result = cls.run_cli_command(
            "scaffold", "connection", "my_connection_1", cwd=cls._get_cwd()
        )
        assert result.exit_code == 0
        result = cls.run_cli_command(
            "scaffold", "connection", "my_connection_2", cwd=cls._get_cwd()
        )
        assert result.exit_code == 0

        config_path_1 = (
            Path(cls._get_cwd())
            / "connections"
            / "my_connection_1"
            / DEFAULT_CONNECTION_CONFIG_FILE
        )

        cls._write_dependencies(
            {"this_is_a_test_dependency": {"version": "==0.1.0"}}, config_path_1
        )

        config_path_2 = (
            Path(cls._get_cwd())
            / "connections"
            / "my_connection_2"
            / DEFAULT_CONNECTION_CONFIG_FILE
        )

        cls._write_dependencies(
            {"this_is_a_test_dependency": {"version": "==0.2.0"}}, config_path_2
        )

    @classmethod
    def _write_dependencies(cls, dependency_dict: Dict, path_to_config: Path):
        """Write a dependency to a configuration file."""
        with path_to_config.open() as fp:
            config = yaml.safe_load(fp)
        config.setdefault("dependencies", {}).update(dependency_dict)
        with path_to_config.open(mode="w") as fp:
            yaml.safe_dump(config, fp)

    def test_error(self):
        """Assert an error occurs."""
        with pytest.raises(
            ClickException,
            match="Error while merging dependencies for connections; Joint version specifier is unsatisfiable for following dependencies",
        ):
            self.run_cli_command("install", cwd=self._get_cwd())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Constants, utility functions and base classes for ACN p2p_libp2p tests"""

import functools
import inspect
import itertools
import platform
import tempfile
from typing import Any, Callable, Type
from unittest import mock

import pytest

from packages.fetchai.protocols.default.message import DefaultMessage

TIMEOUT = 20
TEMP_LIBP2P_TEST_DIR = tempfile.mkdtemp()
ports = itertools.count(10234)

MockDefaultMessageProtocol = mock.Mock()
MockDefaultMessageProtocol.protocol_id = DefaultMessage.protocol_id
MockDefaultMessageProtocol.protocol_specification_id = (
    DefaultMessage.protocol_specification_id
)

SKIP_WINDOWS = pytest.mark.skipif(
    condition=(platform.system() == "Windows"),
    reason="https://github.com/golang/go/issues/51007",
)


def libp2p_log_on_failure(fn: Callable) -> Callable:
    """Decorate a method running a libp2p node to print its logs in case test fails."""

    @functools.wraps(fn)
    def wrapper(self, *args: Any, **kwargs: Any) -> None:  # type: ignore
        try:
            return fn(self, *args, **kwargs)
        except Exception:
            for log_file in getattr(self, "log_files", []):
                print(f"libp2p log file ======================= {log_file}")
                try:
                    with open(log_file, "r") as f:
                        print(f.read())
                except FileNotFoundError:
                    print("FileNotFoundError")
                print("=======================================")
            raise

    return wrapper


def libp2p_log_on_failure_all(cls: Type) -> Type:
    """Wrap test methods to print libp2p logs on failure."""

    def _wrap(func: Callable) -> Callable:
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def _async_wrapped(*args: Any, **kwargs: Any) -> Any:
                return await func(*args, **kwargs)

            return _async_wrapped

        @functools.wraps(func)
        def _sync_wrapped(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        return _sync_wrapped

    for name, value in list(vars(cls).items()):
        if name.startswith("test_") and callable(value):
            setattr(cls, name, _wrap(value))

    return cls



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Check amount of time for acn connection start."""

import os
from pathlib import Path
from typing import Optional, Sequence, Union
from unittest.mock import patch

from aea.configurations.base import ConnectionConfig
from aea.configurations.constants import DEFAULT_LEDGER
from aea.crypto.base import Crypto
from aea.crypto.registries import make_crypto
from aea.crypto.wallet import CryptoStore
from aea.helpers.base import CertRequest, SimpleId
from aea.helpers.multiaddr.base import MultiAddr
from aea.identity.base import Identity

from packages.valory.connections.p2p_libp2p.check_dependencies import build_node
from packages.valory.connections.p2p_libp2p.connection import (
    P2PLibp2pConnection,
    POR_DEFAULT_SERVICE_ID,
)
from packages.valory.connections.p2p_libp2p.consts import (
    LIBP2P_CERT_NOT_AFTER,
    LIBP2P_CERT_NOT_BEFORE,
    LIBP2P_NODE_MODULE_NAME,
)
from packages.valory.connections.p2p_libp2p_client.connection import (
    P2PLibp2pClientConnection,
)
from packages.valory.connections.p2p_libp2p_mailbox.connection import (
    P2PLibp2pMailboxConnection,
)

DEFAULT_DELEGATE_PORT = 11234
DEFAULT_MAILBOX_PORT = 8888
DEFAULT_NODE_PORT = 10234


def _process_cert(key: Crypto, cert: CertRequest, path_prefix: str):
    # must match aea/cli/issue_certificates.py:_process_certificate
    assert cert.public_key is not None
    message = cert.get_message(cert.public_key)
    signature = key.sign_message(message).encode("ascii").hex()
    Path(cert.get_absolute_save_path(path_prefix)).write_bytes(
        signature.encode("ascii")
    )


def _make_libp2p_connection(
    data_dir: str,
    port: int = DEFAULT_NODE_PORT,
    host: str = "127.0.0.1",
    relay: bool = True,
    delegate: bool = False,
    mailbox: bool = False,
    entry_peers: Optional[Sequence[MultiAddr]] = None,
    delegate_port: int = DEFAULT_DELEGATE_PORT,
    delegate_host: str = "127.0.0.1",
    mailbox_port: int = DEFAULT_MAILBOX_PORT,
    mailbox_host: str = "127.0.0.1",
    node_key_file: Optional[str] = None,
    agent_key: Optional[Crypto] = None,
    build_directory: Optional[str] = None,
    peer_registration_delay: str = "0.0",
) -> P2PLibp2pConnection:
    if not os.path.isdir(data_dir) or not os.path.exists(data_dir):
        raise ValueError("Data dir must be directory and exist!")
    log_file = os.path.join(data_dir, "libp2p_node_{}.log".format(port))
    if os.path.exists(log_file):
        os.remove(log_file)
    key = agent_key
    if key is None:
        key = make_crypto(DEFAULT_LEDGER)
    identity = Identity("identity", address=key.address, public_key=key.public_key)
    conn_crypto_store = None
    if node_key_file is not None:
        conn_crypto_store = CryptoStore({DEFAULT_LEDGER: node_key_file})
    else:
        node_key = make_crypto(DEFAULT_LEDGER)
        node_key_path = os.path.join(data_dir, f"{node_key.public_key}.txt")
        node_key.dump(node_key_path)
        conn_crypto_store = CryptoStore({DEFAULT_LEDGER: node_key_path})
    cert_request = CertRequest(
        conn_crypto_store.public_keys[DEFAULT_LEDGER],
        POR_DEFAULT_SERVICE_ID,
        key.identifier,
        LIBP2P_CERT_NOT_BEFORE,
        LIBP2P_CERT_NOT_AFTER,
        "{public_key}",
        f"./{key.address}_cert.txt",
    )
    _process_cert(key, cert_request, path_prefix=data_dir)
    if not build_directory:
        build_directory = os.getcwd()
    if relay and delegate:
        configuration = ConnectionConfig(
            node_key_file=node_key_file,
            local_uri="{}:{}".format(host, port),
            public_uri="{}:{}".format(host, port),
            entry_peers=entry_peers,
            log_file=log_file,
            delegate_uri="{}:{}".format(delegate_host, delegate_port),
            peer_registration_delay=peer_registration_delay,
            connection_id=P2PLibp2pConnection.connection_id,
            build_directory=build_directory,
            cert_requests=[cert_request],
        )
    elif relay and not delegate:
        configuration = ConnectionConfig(
            node_key_file=node_key_file,
            local_uri="{}:{}".format(host, port),
            public_uri="{}:{}".format(host, port),
            entry_peers=entry_peers,
            log_file=log_file,
            peer_registration_delay=peer_registration_delay,
            connection_id=P2PLibp2pConnection.connection_id,
            build_directory=build_directory,
            cert_requests=[cert_request],
        )
    else:
        configuration = ConnectionConfig(
            node_key_file=node_key_file,
            local_uri="{}:{}".format(host, port),
            entry_peers=entry_peers,
            log_file=log_file,
            peer_registration_delay=peer_registration_delay,
            connection_id=P2PLibp2pConnection.connection_id,
            build_directory=build_directory,
            cert_requests=[cert_request],
        )

    if mailbox:
        configuration.config["mailbox_uri"] = f"{mailbox_host}:{mailbox_port}"
    else:
        configuration.config["mailbox_uri"] = ""

    if not os.path.exists(os.path.join(build_directory, LIBP2P_NODE_MODULE_NAME)):
        with patch("builtins.print"):
            build_node(build_directory)
    connection = P2PLibp2pConnection(
        configuration=configuration,
        data_dir=data_dir,
        identity=identity,
        crypto_store=conn_crypto_store,
    )
    return connection


def _make_libp2p_client_connection(
    peer_public_key: str,
    data_dir: str,
    node_port: int = DEFAULT_DELEGATE_PORT,
    node_host: str = "127.0.0.1",
    uri: Optional[str] = None,
    ledger_api_id: Union[SimpleId, str] = DEFAULT_LEDGER,
) -> P2PLibp2pClientConnection:
    if not os.path.isdir(data_dir) or not os.path.exists(data_dir):
        raise ValueError("Data dir must be directory and exist!")
    crypto = make_crypto(ledger_api_id)
    identity = Identity(
        "identity", address=crypto.address, public_key=crypto.public_key
    )
    cert_request = CertRequest(
        peer_public_key,
        POR_DEFAULT_SERVICE_ID,
        ledger_api_id,
        LIBP2P_CERT_NOT_BEFORE,
        LIBP2P_CERT_NOT_AFTER,
        "{public_key}",
        f"./{crypto.address}_cert.txt",
    )
    _process_cert(crypto, cert_request, path_prefix=data_dir)
    configuration = ConnectionConfig(
        tcp_key_file=None,
        nodes=[
            {
                "uri": (
                    str(uri)
                    if uri is not None
                    else "{}:{}".format(node_host, node_port)
                ),
                "public_key": peer_public_key,
            },
        ],
        connection_id=P2PLibp2pClientConnection.connection_id,
        cert_requests=[cert_request],
    )
    return P2PLibp2pClientConnection(
        configuration=configuration, data_dir=data_dir, identity=identity
    )


def _make_libp2p_mailbox_connection(
    peer_public_key: str,
    data_dir: str,
    node_port: int = DEFAULT_MAILBOX_PORT,
    node_host: str = "127.0.0.1",
    uri: Optional[str] = None,
    ledger_api_id: Union[SimpleId, str] = DEFAULT_LEDGER,
) -> P2PLibp2pMailboxConnection:
    if not os.path.isdir(data_dir) or not os.path.exists(data_dir):
        raise ValueError("Data dir must be directory and exist!")
    crypto = make_crypto(ledger_api_id)
    identity = Identity(
        "identity", address=crypto.address, public_key=crypto.public_key
    )
    cert_request = CertRequest(
        peer_public_key,
        POR_DEFAULT_SERVICE_ID,
        ledger_api_id,
        LIBP2P_CERT_NOT_BEFORE,
        LIBP2P_CERT_NOT_AFTER,
        "{public_key}",
        f"./{crypto.address}_cert.txt",
    )
    _process_cert(crypto, cert_request, path_prefix=data_dir)
    configuration = ConnectionConfig(
        tcp_key_file=None,
        nodes=[
            {
                "uri": (
                    str(uri)
                    if uri is not None
                    else "{}:{}".format(node_host, node_port)
                ),
                "public_key": peer_public_key,
            },
        ],
        connection_id=P2PLibp2pMailboxConnection.connection_id,
        cert_requests=[cert_request],
    )
    return P2PLibp2pMailboxConnection(
        configuration=configuration, data_dir=data_dir, identity=identity
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Constants."""

import os
from pathlib import Path

from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_ethereum.ethereum import (
    DEFAULT_EIP1559_STRATEGY,
    DEFAULT_GAS_STATION_STRATEGY,
)

from aea.configurations.constants import PRIVATE_KEY_PATH_SCHEMA
from aea.crypto.ledger_apis import ETHEREUM_DEFAULT_ADDRESS, ETHEREUM_DEFAULT_CHAIN_ID

DATA_DIR = Path(__file__).parent / "data"

ETHEREUM_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(EthereumCrypto.identifier)
ETHEREUM_PRIVATE_KEY_TWO_FILE = "ethereum_private_key_two.txt"

ETHEREUM_ADDRESS_ONE = "0x46F415F7BF30f4227F98def9d2B22ff62738fD68"
ETHEREUM_ADDRESS_TWO = "0x7A1236d5195e31f1F573AD618b2b6FEFC85C5Ce6"

ETHEREUM_PRIVATE_KEY_PATH = os.path.join(DATA_DIR, ETHEREUM_PRIVATE_KEY_FILE)
ETHEREUM_PRIVATE_KEY_TWO_PATH = os.path.join(DATA_DIR, ETHEREUM_PRIVATE_KEY_TWO_FILE)

ETHEREUM_TESTNET_CONFIG = {
    "address": ETHEREUM_DEFAULT_ADDRESS,
    "chain_id": ETHEREUM_DEFAULT_CHAIN_ID,
    "default_gas_price_strategy": "gas_station",
    "gas_price_strategies": {
        "gas_station": DEFAULT_GAS_STATION_STRATEGY,
        "eip1559": DEFAULT_EIP1559_STRATEGY,
    },
}

FUNDED_ETH_PRIVATE_KEY_1 = (
    "0xa337a9149b4e1eafd6c21c421254cf7f98130233595db25f0f6f0a545fb08883"
)
FUNDED_ETH_PRIVATE_KEY_2 = (
    "0x04b4cecf78288f2ab09d1b4c60219556928f86220f0fb2dcfc05e6a1c1149dbf"
)
FUNDED_ETH_PRIVATE_KEY_3 = (
    "0x6F611408F7EF304947621C51A4B7D84A13A2B9786E9F984DA790A096E8260C64"
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Module to declare constants."""

from pathlib import Path
from typing import Dict, List

AEA_DIR = Path(__file__).parent.parent
_FETCHAI_IDENTIFIER = "fetchai"
_ETHEREUM_IDENTIFIER = "ethereum"
_COSMOS_IDENTIFIER = "cosmos"
_SOLANA_IDENTIFIER = "solana"
SIGNING_PROTOCOL = "open_aea/signing:latest"
SIGNING_PROTOCOL_WITH_HASH = (
    "open_aea/signing:1.0.0:bafybeifsjmldwyki3beqyvdt5lzenrg6wyrqaar5plc5rpnvtc4zlentye"
)
DEFAULT_LEDGER = _ETHEREUM_IDENTIFIER
PRIVATE_KEY_PATH_SCHEMA = "{}_private_key.txt"
MULTIKEY_FILENAME = "keys.json"
PRIVATE_KEY = "private_key"
ADDRESS = "address"
LEDGER = "ledger"
DEFAULT_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(DEFAULT_LEDGER)
DEFAULT_LICENSE = "Apache-2.0"
DISTRIBUTED_PACKAGES: List[str] = []
DEFAULT_SEARCH_SERVICE_ADDRESS = "open_aea/soef:any"
DEFAULT_INPUT_FILE_NAME = "./input_file"
DEFAULT_OUTPUT_FILE_NAME = "./output_file"
SCAFFOLD_PUBLIC_ID = "open_aea/scaffold:0.1.0"
PACKAGES = "packages"
DEFAULT_REGISTRY_NAME = PACKAGES
VENDOR = "vendor"
AGENT = "agent"
AGENTS = "agents"
BUILD = "build"
CUSTOM = "custom"
CUSTOMS = "customs"
CONNECTION = "connection"
CONNECTIONS = "connections"
CONTRACT = "contract"
CONTRACTS = "contracts"
PROTOCOL = "protocol"
PROTOCOLS = "protocols"
SKILL = "skill"
SKILLS = "skills"
SERVICE = "service"
SERVICES = "services"
DEFAULT_README_FILE = "README.md"
DEFAULT_VERSION = "0.1.0"
DEFAULT_AEA_CONFIG_FILE = "aea-config.yaml"
DEFAULT_SKILL_CONFIG_FILE = "skill.yaml"
DEFAULT_CONNECTION_CONFIG_FILE = "connection.yaml"
DEFAULT_CUSTOM_COMPONENT_CONFIG_FILE = "component.yaml"
DEFAULT_CONTRACT_CONFIG_FILE = "contract.yaml"
DEFAULT_PROTOCOL_CONFIG_FILE = "protocol.yaml"
DEFAULT_SERVICE_CONFIG_FILE = "service.yaml"
PACKAGE_PUBLIC_ID_VAR_NAME = "PUBLIC_ID"
DEFAULT_FINGERPRINT_IGNORE_PATTERNS = [
    ".DS_Store",
    "*__pycache__/*",
    "*__pycache__",
    "*.pyc",
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_SERVICE_CONFIG_FILE,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_CUSTOM_COMPONENT_CONFIG_FILE,
    PRIVATE_KEY_PATH_SCHEMA.format("*"),
]
DEFAULT_PYPI_INDEX_URL = "https://pypi.org/simple"
DEFAULT_GIT_REF = "master"
DEFAULT_LOGGING_CONFIG = {"version": 1, "disable_existing_loggers": False}
IMPORT_TEMPLATE_1 = "from packages.{author}.{type}.{name}"
IMPORT_TEMPLATE_2 = "import packages.{author}.{type}.{name}"
DEFAULT_ENV_DOTFILE = ".env"
DOTTED_PATH_MODULE_ELEMENT_SEPARATOR = ":"
DEFAULT_BUILD_DIR_NAME = ".build"
DEFAULT_DEPENDENCIES: Dict[str, Dict] = {"open-aea-ledger-ethereum": {}}

CONFIG_FILE_TO_PACKAGE_TYPE = {
    DEFAULT_SKILL_CONFIG_FILE: SKILL,
    DEFAULT_PROTOCOL_CONFIG_FILE: PROTOCOL,
    DEFAULT_CONNECTION_CONFIG_FILE: CONNECTION,
    DEFAULT_CONTRACT_CONFIG_FILE: CONTRACT,
    DEFAULT_SERVICE_CONFIG_FILE: SERVICE,
    DEFAULT_AEA_CONFIG_FILE: AGENT,
    DEFAULT_CUSTOM_COMPONENT_CONFIG_FILE: CUSTOM,
}  # type: Dict[str, str]

PACKAGE_TYPE_TO_CONFIG_FILE = {
    value: key for key, value in CONFIG_FILE_TO_PACKAGE_TYPE.items()
}  # type: Dict[str, str]

CRYPTO_PLUGIN_GROUP = "aea.cryptos"
LEDGER_APIS_PLUGIN_GROUP = "aea.ledger_apis"
FAUCET_APIS_PLUGIN_GROUP = "aea.faucet_apis"
ALLOWED_GROUPS = {
    CRYPTO_PLUGIN_GROUP,
    LEDGER_APIS_PLUGIN_GROUP,
    FAUCET_APIS_PLUGIN_GROUP,
}
AEA_MANAGER_DATA_DIRNAME = "data"
LAUNCH_SUCCEED_MESSAGE = "Start processing messages..."

PROTOCOL_LANGUAGE_PYTHON = "python"
PROTOCOL_LANGUAGE_GO = "go"
PROTOCOL_LANGUAGE_CPP = "cpp"
PROTOCOL_LANGUAGE_JAVA = "java"
PROTOCOL_LANGUAGE_CSHARP = "csharp"
PROTOCOL_LANGUAGE_RUBY = "ruby"
PROTOCOL_LANGUAGE_OBJC = "objc"
PROTOCOL_LANGUAGE_JS = "js"
SUPPORTED_PROTOCOL_LANGUAGES = [
    PROTOCOL_LANGUAGE_PYTHON,
    PROTOCOL_LANGUAGE_GO,
    PROTOCOL_LANGUAGE_CPP,
    PROTOCOL_LANGUAGE_JAVA,
    PROTOCOL_LANGUAGE_CSHARP,
    PROTOCOL_LANGUAGE_RUBY,
    PROTOCOL_LANGUAGE_OBJC,
    PROTOCOL_LANGUAGE_JS,
]
DEFAULT_CERTS_DIR_NAME = ".certs"
DEFAULT_IGNORE_DIRS_AGENT_FINGERPRINT = [
    SKILLS,
    PROTOCOLS,
    CONTRACTS,
    CONNECTIONS,
    VENDOR,
    DEFAULT_BUILD_DIR_NAME,
    DEFAULT_CERTS_DIR_NAME,
]

ITEM_TYPE_TO_PLURAL = {
    PROTOCOL: PROTOCOLS,
    AGENT: AGENTS,
    CONTRACT: CONTRACTS,
    CONNECTION: CONNECTIONS,
    SKILL: SKILLS,
    CUSTOM: CUSTOMS,
}

ITEM_TYPE_PLURAL_TO_TYPE = {v: k for k, v in ITEM_TYPE_TO_PLURAL.items()}
SCAFFOLD_PACKAGES = [
    AEA_DIR / "protocols" / "scaffold",
    AEA_DIR / "connections" / "scaffold",
    AEA_DIR / "contracts" / "scaffold",
    AEA_DIR / "skills" / "scaffold",
]
PYCACHE = "__pycache__"
AEA_TEST_DIRNAME = "tests"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for `aea push-all` command."""

from collections import OrderedDict
from pathlib import Path
from unittest import mock

import click
import pytest

from aea.cli.registry.settings import REMOTE_HTTP, REMOTE_IPFS
from aea.configurations.constants import PACKAGES
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.package_manager.v0 import PackageManagerV0
from aea.package_manager.v1 import PackageManagerV1
from aea.test_tools.test_cases import BaseAEATestCase

TEST_PACKAGE = PublicId(
    "open_aea",
    "signing",
    "0.1.0",
    "bafybeiambqptflge33eemdhis2whik67hjplfnqwieoa6wblzlaf7vuo44",
)
TEST_PACKAGE_ID = PackageId(PackageType.PROTOCOL, TEST_PACKAGE)
TEST_PACKAGES = [
    (
        TEST_PACKAGE_ID,
        TEST_PACKAGE.hash,
    )
]


class TestPushAll(BaseAEATestCase):
    """Test `push-all` command"""

    @pytest.mark.parametrize(
        "package_manager",
        (
            PackageManagerV1(
                path=Path(".", PACKAGES),
                dev_packages=OrderedDict(TEST_PACKAGES),
            ),
            PackageManagerV0(
                path=Path(".", PACKAGES),
                packages=OrderedDict(TEST_PACKAGES),
            ),
        ),
    )
    def test_run(self, package_manager) -> None:
        """Test command invocation"""
        with mock.patch("aea.cli.push_all.push_item_ipfs"), mock.patch(
            "aea.cli.push_all.get_default_remote_registry", return_value=REMOTE_IPFS
        ), mock.patch(
            "aea.cli.push_all.get_package_manager", return_value=package_manager
        ):
            result = self.run_cli_command("push-all", "--remote")

            package_path = package_manager.package_path_from_package_id(TEST_PACKAGE_ID)
            assert result.exit_code == 0, result.output
            assert f"Pushing: {package_path}" in result.output

    def test_local(
        self,
    ) -> None:
        """Test"""

        with pytest.raises(
            click.ClickException,
            match="Pushing all packages is not supported for the local registry",
        ):
            self.run_cli_command("push-all", "--local")

    def test_remote_http(
        self,
    ) -> None:
        """Test command invocation"""

        with mock.patch(
            "aea.cli.push_all.get_default_remote_registry", return_value=REMOTE_HTTP
        ):
            with pytest.raises(
                click.ClickException,
                match="Pushing all packages is not supported for the HTTP registry",
            ):
                self.run_cli_command("push-all", "--remote")

    def test_retries_applied(self) -> None:
        """Test retries flag works for pushing packages."""
        with mock.patch(
            "aea.cli.push_all.get_package_manager",
            return_value=PackageManagerV1(
                path=Path(".", PACKAGES),
                dev_packages=OrderedDict(TEST_PACKAGES[:1]),
            ),
        ), mock.patch(
            "aea.cli.push_all.get_default_remote_registry", return_value=REMOTE_IPFS
        ), mock.patch(
            "pathlib.Path.glob", return_value=[]
        ):
            with mock.patch(
                "aea.cli.push.IPFSTool.add", side_effect=Exception("expected")
            ) as push_item_ipfs_mock, pytest.raises(Exception, match="expected"):
                self.run_cli_command("push-all", "--remote")
            push_item_ipfs_mock.assert_called_once()

            with mock.patch(
                "aea.cli.push.IPFSTool.add", side_effect=Exception("expected")
            ) as push_item_ipfs_mock, pytest.raises(Exception, match="expected"):
                self.run_cli_command("push-all", "--remote", "--retries=3")
            assert push_item_ipfs_mock.call_count == 3



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains Negative tests for Libp2p connection."""

import asyncio
from asyncio.futures import Future
from unittest.mock import Mock, patch

import pytest

from packages.valory.connections.p2p_libp2p.connection import Libp2pNode
from packages.valory.protocols.acn.message import AcnMessage

DEFAULT_NET_SIZE = 4


@pytest.mark.asyncio
async def test_max_restarts() -> None:
    """Test node max restarts exception."""
    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp", max_restarts=0)
    with pytest.raises(ValueError, match="Max restarts attempts reached:"):
        await node.restart()


@pytest.mark.asyncio
async def test_send_acn_confirm_failed() -> None:
    """Test nodeclient send fails on confirmation from other point ."""

    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp")
    f: Future = Future()
    f.set_result(None)
    node.pipe = Mock()
    node.pipe.connect = Mock(return_value=f)
    node.pipe.write = Mock(return_value=f)

    node_client = node.get_client()
    status = Mock()
    status.code = int(AcnMessage.StatusBody.StatusCode.ERROR_GENERIC)
    status_future: Future = Future()
    status_future.set_result(status)
    with patch.object(
        node_client, "make_acn_envelope_message", return_value=b"some_data"
    ), patch.object(
        node_client, "wait_for_status", lambda: status_future
    ), pytest.raises(
        Exception, match=r"failed to send envelope. got error confirmation"
    ):
        await node_client.send_envelope(Mock())


@pytest.mark.asyncio
async def test_send_acn_confirm_timeout() -> None:
    """Test node client send fails on timeout."""

    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp")
    f: Future = Future()
    f.set_result(None)
    node.pipe = Mock()
    node.pipe.connect = Mock(return_value=f)
    node.pipe.write = Mock(return_value=f)

    node_client = node.get_client()
    node_client.ACN_ACK_TIMEOUT = 0.5
    with patch.object(
        node_client, "make_acn_envelope_message", return_value=b"some_data"
    ), patch("asyncio.wait_for", side_effect=asyncio.TimeoutError()), pytest.raises(
        Exception, match=r"acn status await timeout!"
    ):
        await node_client.send_envelope(Mock())


@pytest.mark.asyncio
async def test_acn_decode_error_on_read() -> None:
    """Test ACN decode error on read."""

    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp")
    f: Future = Future()
    f.set_result(b"some_data")
    node.pipe = Mock()
    node.pipe.connect = Mock(return_value=f)

    node_client = node.get_client()
    node_client.ACN_ACK_TIMEOUT = 0.5

    with patch.object(node_client, "_read", lambda: f), patch.object(
        node_client, "write_acn_status_error", return_value=f
    ) as mocked_write_acn_status_error, pytest.raises(
        Exception, match=r"Error parsing acn message:"
    ):
        await node_client.read_envelope()

    mocked_write_acn_status_error.assert_called_once()


@pytest.mark.asyncio
async def test_write_acn_error() -> None:
    """Test write ACN error."""

    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp")
    f: Future = Future()
    f.set_result(b"some_data")
    node.pipe = Mock()
    node.pipe.connect = Mock(return_value=f)

    node_client = node.get_client()

    with patch.object(node_client, "_write", return_value=f) as write_mock:
        await node_client.write_acn_status_error("some error")

    write_mock.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for acn protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.valory.protocols.acn.custom_types import AgentRecord, StatusBody
from packages.valory.protocols.acn.message import AcnMessage


class TestMessageAcn(BaseProtocolMessagesTestCase):
    """Test for the 'acn' protocol message."""

    MESSAGE_CLASS = AcnMessage

    def build_messages(self) -> List[AcnMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            AcnMessage(
                performative=AcnMessage.Performative.REGISTER,
                record=AgentRecord(
                    address="address",
                    public_key="pbk",
                    peer_public_key="peerpbk",
                    signature="sign",
                    service_id="acn",
                    ledger_id="fetchai",
                ),
            ),
            AcnMessage(
                performative=AcnMessage.Performative.LOOKUP_REQUEST,
                agent_address="some str",
            ),
            AcnMessage(
                performative=AcnMessage.Performative.LOOKUP_RESPONSE,
                record=AgentRecord(
                    address="address",
                    public_key="pbk",
                    peer_public_key="peerpbk",
                    signature="sign",
                    service_id="acn",
                    ledger_id="fetchai",
                ),
            ),
            AcnMessage(
                performative=AcnMessage.Performative.AEA_ENVELOPE,
                envelope=b"some_bytes",
                record=AgentRecord(
                    address="address",
                    public_key="pbk",
                    peer_public_key="peerpbk",
                    signature="sign",
                    service_id="acn",
                    ledger_id="fetchai",
                ),
            ),
            AcnMessage(
                performative=AcnMessage.Performative.STATUS,
                body=StatusBody(
                    status_code=AcnMessage.StatusBody.StatusCode.ERROR_UNSUPPORTED_VERSION,
                    msgs=["pbk"],
                ),
            ),
        ]

    def build_inconsistent(self) -> List[AcnMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            AcnMessage(
                performative=AcnMessage.Performative.REGISTER,
                # skip content: record
            ),
            AcnMessage(
                performative=AcnMessage.Performative.LOOKUP_REQUEST,
                # skip content: agent_address
            ),
            AcnMessage(
                performative=AcnMessage.Performative.LOOKUP_RESPONSE,
                # skip content: record
            ),
            AcnMessage(
                performative=AcnMessage.Performative.AEA_ENVELOPE,
                # skip content: envelope
                record=AgentRecord(
                    address="address",
                    public_key="pbk",
                    peer_public_key="peerpbk",
                    signature="sign",
                    service_id="acn",
                    ledger_id="fetchai",
                ),
            ),
            AcnMessage(
                performative=AcnMessage.Performative.STATUS,
                # skip content: body
            ),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the erc1155_deploy skill."""

# pylint: skip-file

import pytest

from aea.exceptions import AEAEnforceError
from aea.protocols.dialogue.base import DialogueLabel
from aea.test_tools.test_skill import COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.erc1155_deploy.dialogues import (
    ContractApiDialogue,
    DefaultDialogue,
    FipaDialogue,
    LedgerApiDialogue,
    OefSearchDialogue,
    SigningDialogue,
)
from packages.fetchai.skills.erc1155_deploy.tests.intermediate_class import (
    ERC1155DeployTestCase,
)
from packages.open_aea.protocols.signing.message import SigningMessage
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage


class TestDialogues(ERC1155DeployTestCase):
    """Test dialogue classes of erc1155_deploy."""

    def test_contract_api_dialogue(self):
        """Test the ContractApiDialogue class."""
        contract_api_dialogue = ContractApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # terms
        with pytest.raises(ValueError, match="Terms not set!"):
            assert contract_api_dialogue.terms
        contract_api_dialogue.terms = self.mocked_terms
        with pytest.raises(AEAEnforceError, match="Terms already set!"):
            contract_api_dialogue.terms = self.mocked_terms
        assert contract_api_dialogue.terms == self.mocked_terms

    def test_contract_api_dialogues(self):
        """Test the ContractApiDialogues class."""
        _, dialogue = self.contract_api_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            callable=self.callable,
            kwargs=self.kwargs,
        )
        assert dialogue.role == ContractApiDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)

    def test_default_dialogues(self):
        """Test the DefaultDialogues class."""
        _, dialogue = self.default_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=DefaultMessage.Performative.BYTES,
            content=b"some_content",
        )
        assert dialogue.role == DefaultDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_fipa_dialogue(self):
        """Test the FipaDialogue class."""
        fipa_dialogue = FipaDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=FipaDialogue.Role.BUYER,
        )

        # proposal
        with pytest.raises(ValueError, match="Proposal not set!"):
            assert fipa_dialogue.proposal
        fipa_dialogue.proposal = self.mocked_registration_description
        with pytest.raises(AEAEnforceError, match="Proposal already set!"):
            fipa_dialogue.proposal = self.mocked_registration_description
        assert fipa_dialogue.proposal == self.mocked_registration_description

    def test_fipa_dialogues(self):
        """Test the FipaDialogues class."""
        _, dialogue = self.fipa_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=FipaMessage.Performative.CFP,
            query=self.mocked_query,
        )
        assert dialogue.role == FipaDialogue.Role.SELLER
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_ledger_api_dialogue(self):
        """Test the LedgerApiDialogue class."""
        ledger_api_dialogue = LedgerApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_signing_dialogue
        with pytest.raises(ValueError, match="Associated signing dialogue not set!"):
            assert ledger_api_dialogue.associated_signing_dialogue
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=SigningDialogue.Role.SKILL,
        )
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue
        with pytest.raises(
            AEAEnforceError, match="Associated signing dialogue already set!"
        ):
            ledger_api_dialogue.associated_signing_dialogue = signing_dialogue
        assert ledger_api_dialogue.associated_signing_dialogue == signing_dialogue

    def test_ledger_api_dialogues(self):
        """Test the LedgerApiDialogues class."""
        _, dialogue = self.ledger_api_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=self.ledger_id,
            address=self.address,
        )
        assert dialogue.role == LedgerApiDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)

    def test_oef_search_dialogues(self):
        """Test the OefSearchDialogues class."""
        _, dialogue = self.oef_search_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=self.mocked_query,
        )
        assert dialogue.role == OefSearchDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)

    def test_signing_dialogue(self):
        """Test the SigningDialogue class."""
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_contract_api_dialogue
        with pytest.raises(
            ValueError, match="Associated contract api dialogue not set!"
        ):
            assert signing_dialogue.associated_contract_api_dialogue
        contract_api_dialogue = ContractApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )
        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        with pytest.raises(
            AEAEnforceError, match="Associated contract api dialogue already set!"
        ):
            signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        assert (
            signing_dialogue.associated_contract_api_dialogue == contract_api_dialogue
        )

    def test_signing_dialogues(self):
        """Test the SigningDialogues class."""
        _, dialogue = self.signing_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            terms=self.mocked_terms,
            raw_transaction=self.mocked_raw_tx,
        )
        assert dialogue.role == SigningDialogue.Role.SKILL
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of background profiling daemon."""

import asyncio
import datetime
import gc
import logging
import platform
import textwrap
import threading
import time
from collections import Counter
from concurrent.futures._base import CancelledError
from functools import wraps
from typing import Any, Callable
from typing import Counter as CounterType
from typing import Dict, List, Tuple, Type

from aea.helpers.async_utils import Runnable
from aea.helpers.profiler_type_black_list import PROFILER_TYPE_BLACK_LIST

BYTES_TO_MBYTES = 1024**-2

lock = threading.Lock()

_default_logger = logging.getLogger(__file__)

if platform.system() == "Windows":  # pragma: nocover
    import win32process  # type: ignore  # pylint: disable=import-error,import-outside-toplevel

    WIN32_PROCESS_TIMES_TICKS_PER_SECOND = 1e7

    def get_current_process_memory_usage() -> float:
        """Get current process memory usage in MB."""
        d = win32process.GetProcessMemoryInfo(win32process.GetCurrentProcess())  # type: ignore
        return float(d["WorkingSetSize"]) * BYTES_TO_MBYTES

    def get_peak_process_memory_usage() -> float:
        """Get current process memory usage in MB."""
        d = win32process.GetProcessMemoryInfo(win32process.GetCurrentProcess())  # type: ignore
        return float(d["PeakWorkingSetSize"]) * BYTES_TO_MBYTES

    def get_current_process_cpu_time() -> float:
        """Get current process cpu time in seconds."""
        d = win32process.GetProcessTimes(win32process.GetCurrentProcess())  # type: ignore
        return d["UserTime"] / WIN32_PROCESS_TIMES_TICKS_PER_SECOND

else:
    import resource
    import tracemalloc

    def get_current_process_memory_usage() -> float:
        """Get current process memory usage in MB."""
        return tracemalloc.get_traced_memory()[0] * BYTES_TO_MBYTES

    def get_peak_process_memory_usage() -> float:
        """Get current process memory usage in MB."""
        return tracemalloc.get_traced_memory()[1] * BYTES_TO_MBYTES

    def get_current_process_cpu_time() -> float:
        """Get current process cpu time in seconds."""
        return resource.getrusage(resource.RUSAGE_SELF).ru_utime


class Profiling(Runnable):
    """Profiling service."""

    def __init__(
        self,
        types_to_track: List[Type],
        period: int = 0,
        output_function: Callable[[str], None] = lambda x: print(x, flush=True),
    ) -> None:
        """
        Init profiler.

        :param period: delay between profiling output in seconds.
        :param types_to_track: object types to count
        :param output_function: function to display output, one str argument.
        """
        if period < 1:  # pragma: nocover
            raise ValueError("Profiling frequency should be at least 1 second!")
        super().__init__(threaded=True)
        self._period = period
        self._start_ts = time.time()
        self._types_to_track: List[Type] = types_to_track
        self._output_function = output_function
        self.object_counts: Dict[Type, List[int]] = {
            obj: [0, 0] for obj in types_to_track
        }  # {object: [instances_created, instances_deleted]}
        self.set_counters()

        if platform.system() != "Windows":
            tracemalloc.start()

    def set_counters(self) -> None:
        """Modify __new__ and __del__ to count objects created created and destroyed."""

        def call_count(wrapped: Callable, index: int, obj: Any) -> Callable:
            orig_new = obj.__new__

            @wraps(wrapped)
            def wrapper(*args: Any, **kwargs: Any) -> Callable:
                self.object_counts[obj][index] += 1
                # Avoid TypeError: object.__new__() takes exactly one argument
                if orig_new is object.__new__:
                    return orig_new(args[0])  # pragma: nocover
                return wrapped(*args, **kwargs)

            return wrapper

        for t in self._types_to_track:
            t.__new__ = call_count(t.__new__, 0, t)
            # For some reason, if we don't init the __del__ method with an empty function, the next
            # line will raise the exception: AttributeError: type object 'Message' has no attribute '__del__'
            t.__del__ = lambda _: None
            t.__del__ = call_count(t.__del__, 1, t)

    async def run(self) -> None:
        """Run profiling."""
        try:
            while True:
                await asyncio.sleep(self._period)
                self.output_profile_data()
        except CancelledError:  # pragma: nocover
            pass
        except Exception:  # pragma: nocover
            _default_logger.exception("Exception in Profiling")
            raise
        finally:
            if platform.system() != "Windows":
                tracemalloc.stop()

    def output_profile_data(self) -> None:
        """Render profiling data and call output_function."""
        data = self.get_profile_data()
        text = (
            textwrap.dedent(f"""
        Profiling details for current AEA process: {datetime.datetime.now()}
        =============================================
        Run time: {data["run_time"]:.6f} seconds
        Cpu time: {data["cpu_time"]:.6f} seconds,
        Cpu/Run time: {100*data["cpu_time"]/data["run_time"]:.6f}%
        Memory: {data["mem"]:.6f} MB [Peak {data["mem_peak"]:.6f} MB]
        Threads: {data["threads"]['amount']}  {data["threads"]['names']}
        Objects present:
        """)
            + "\n".join(
                [
                    f" * {i.__name__} (present):  {c}"
                    for i, c in data["objects_present"].items()
                ]
            )
            + "\n"
            + """Objects created:\n"""
            + "\n".join(
                [
                    f" * {i.__name__} (created):  {c}"
                    for i, c in data["objects_created"].items()
                ]
            )
            + "\n"
            + """Most common objects in garbage collector (excluding blacklisted):\n"""
            + "\n".join(
                [f" * {i[0]} (gc):  {i[1]}" for i in data["most_common_objects_in_gc"]]
            )
            + "\n"
        )
        self._output_function(text)

    def get_profile_data(self) -> Dict:
        """Get profiling data dict."""
        return {
            "run_time": time.time() - self._start_ts,
            "cpu_time": get_current_process_cpu_time(),
            "mem": get_current_process_memory_usage(),
            "mem_peak": get_peak_process_memory_usage(),
            "threads": {
                "amount": threading.active_count(),
                "names": [i.name for i in threading.enumerate()],
            },
            "objects_present": {k: v[0] - v[1] for k, v in self.object_counts.items()},
            "objects_created": {k: v[0] for k, v in self.object_counts.items()},
            "most_common_objects_in_gc": get_most_common_objects_in_gc(),
        }


def get_most_common_objects_in_gc(number: int = 15) -> List[Tuple[str, int]]:
    """Get the highest-count objects in the garbage collector."""

    object_count: CounterType = Counter()
    lock.acquire()
    try:
        for obj in gc.get_objects():
            object_type = str(
                getattr(obj, "__class__", type(obj)).__name__
            )  # not all objects have the __class__ attribute
            if object_type not in PROFILER_TYPE_BLACK_LIST:
                object_count[object_type] += 1
    finally:
        lock.release()
    return object_count.most_common(number)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the generic buyer skill."""

# pylint: skip-file

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

import pytest

from aea.crypto.ledger_apis import LedgerApis
from aea.helpers.search.models import Description
from aea.helpers.transaction.base import (
    RawTransaction,
    SignedTransaction,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.generic_buyer.behaviours import GenericTransactionBehaviour
from packages.fetchai.skills.generic_buyer.dialogues import (
    FipaDialogue,
    FipaDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    OefSearchDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.fetchai.skills.generic_buyer.handlers import (
    GenericFipaHandler,
    GenericLedgerApiHandler,
    GenericOefSearchHandler,
    GenericSigningHandler,
    LEDGER_API_ADDRESS,
)
from packages.fetchai.skills.generic_buyer.strategy import GenericStrategy
from packages.open_aea.protocols.signing.message import SigningMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

PACKAGE_ROOT = Path(__file__).parent.parent


class TestGenericFipaHandler(BaseSkillTestCase):
    """Test fipa handler of generic buyer."""

    path_to_skill = PACKAGE_ROOT

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.fipa_handler = cast(
            GenericFipaHandler, cls._skill.skill_context.handlers.fipa
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.list_of_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )

    def test_setup(self):
        """Test the setup method of the fipa handler."""
        assert self.fipa_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the fipa handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=FipaMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=FipaMessage.Performative.ACCEPT,
        )

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid fipa message={incoming_message}, unidentified dialogue.",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=DefaultMessage,
            performative=DefaultMessage.Performative.ERROR,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            error_code=DefaultMessage.ErrorCode.INVALID_DIALOGUE,
            error_msg="Invalid dialogue.",
            error_data={"fipa_message": incoming_message.encode()},
        )
        assert has_attributes, error_str

    def test_handle_propose_is_affordable_and_is_acceptable(self):
        """Test the _handle_propose method of the fipa handler."""
        # setup
        proposal = Description(
            {
                "ledger_id": self.strategy.ledger_id,
                "price": 100,
                "currency_id": "FET",
                "service_id": "some_service_id",
                "quantity": 1,
                "tx_nonce": "some_tx_nonce",
            }
        )
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.PROPOSE,
            proposal=proposal,
        )

        # operation
        with patch.object(
            self.strategy,
            "is_acceptable_proposal",
            return_value=True,
        ):
            with patch.object(
                self.strategy,
                "is_affordable_proposal",
                return_value=True,
            ):
                with patch.object(
                    self.fipa_handler.context.logger, "log"
                ) as mock_logger:
                    self.fipa_handler.handle(incoming_message)

        # after
        incoming_message = cast(FipaMessage, incoming_message)
        mock_logger.assert_any_call(
            logging.INFO,
            f"received proposal={incoming_message.proposal.values} from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )
        mock_logger.assert_any_call(
            logging.INFO,
            f"accepting the proposal from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=FipaMessage,
            performative=FipaMessage.Performative.ACCEPT,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
        )
        assert has_attributes, error_str

    def test_handle_propose_not_is_affordable_or_not_is_acceptable(self):
        """Test the _handle_propose method of the fipa handler."""
        # setup
        proposal = Description(
            {
                "ledger_id": self.strategy.ledger_id,
                "price": 100,
                "currency_id": "FET",
                "service_id": "some_service_id",
                "quantity": 1,
                "tx_nonce": "some_tx_nonce",
            }
        )
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.PROPOSE,
            proposal=proposal,
        )

        # operation
        with patch.object(
            self.strategy,
            "is_acceptable_proposal",
            return_value=False,
        ):
            with patch.object(
                self.strategy,
                "is_affordable_proposal",
                return_value=False,
            ):
                with patch.object(
                    self.fipa_handler.context.logger, "log"
                ) as mock_logger:
                    self.fipa_handler.handle(incoming_message)

        # after
        incoming_message = cast(FipaMessage, incoming_message)
        mock_logger.assert_any_call(
            logging.INFO,
            f"received proposal={incoming_message.proposal.values} from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )
        mock_logger.assert_any_call(
            logging.INFO,
            f"declining the proposal from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=FipaMessage,
            performative=FipaMessage.Performative.DECLINE,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
        )
        assert has_attributes, error_str

    def test_handle_decline_decline_cfp(self):
        """Test the _handle_decline method of the fipa handler where the end state is decline_cfp."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.DECLINE,
        )

        # before
        for (
            end_state_numbers
        ) in self.fipa_dialogues.dialogue_stats.self_initiated.values():
            assert end_state_numbers == 0
        for (
            end_state_numbers
        ) in self.fipa_dialogues.dialogue_stats.other_initiated.values():
            assert end_state_numbers == 0

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received DECLINE from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )

        for (
            end_state_numbers
        ) in self.fipa_dialogues.dialogue_stats.other_initiated.values():
            assert end_state_numbers == 0
        for (
            end_state,
            end_state_numbers,
        ) in self.fipa_dialogues.dialogue_stats.self_initiated.items():
            if end_state == FipaDialogue.EndState.DECLINED_CFP:
                assert end_state_numbers == 1
            else:
                assert end_state_numbers == 0

    def test_handle_decline_decline_accept(self):
        """Test the _handle_decline method of the fipa handler where the end state is decline_accept."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:3],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.DECLINE,
        )

        # before
        for end_state_numbers in list(
            self.fipa_dialogues.dialogue_stats.self_initiated.values()
        ) + list(self.fipa_dialogues.dialogue_stats.other_initiated.values()):
            assert end_state_numbers == 0

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received DECLINE from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )

        for (
            end_state_numbers
        ) in self.fipa_dialogues.dialogue_stats.other_initiated.values():
            assert end_state_numbers == 0
        for (
            end_state,
            end_state_numbers,
        ) in self.fipa_dialogues.dialogue_stats.self_initiated.items():
            if end_state == FipaDialogue.EndState.DECLINED_ACCEPT:
                assert end_state_numbers == 1
            else:
                assert end_state_numbers == 0

    def test_handle_match_accept_is_ledger_tx(self):
        """Test the _handle_match_accept method of the fipa handler where is_ledger_tx is True."""
        # setup
        self.strategy._is_ledger_tx = True

        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:3],
        )
        fipa_dialogue.terms = Terms(
            "some_ledger_id",
            self.skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        incoming_message = cast(
            FipaMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=fipa_dialogue,
                performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                info={"info": {"address": "some_term_sender_address"}},
            ),
        )

        # operation
        with patch.object(
            self.fipa_handler.context.logger, "log"
        ) as mock_logger_handler:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger_handler.assert_any_call(
            logging.INFO,
            f"received MATCH_ACCEPT_W_INFORM from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]} with info={incoming_message.info}",
        )

        # operation
        with patch.object(
            self.fipa_handler.context.behaviours.transaction.context.logger, "log"
        ) as _:
            self.fipa_handler.context.behaviours.transaction.act()

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            terms=fipa_dialogue.terms,
        )
        assert has_attributes, error_str

    def test_handle_match_accept_not_is_ledger_tx(self):
        """Test the _handle_match_accept method of the fipa handler where is_ledger_tx is False."""
        # setup
        self.strategy._is_ledger_tx = False

        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:3],
        )
        incoming_message = cast(
            FipaMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=fipa_dialogue,
                performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                info={"info": {"address": "some_term_sender_address"}},
            ),
        )

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received MATCH_ACCEPT_W_INFORM from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]} with info={incoming_message.info}",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=FipaMessage,
            performative=FipaMessage.Performative.INFORM,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
            info={"Done": "Sending payment via bank transfer"},
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO,
            f"informing counterparty={COUNTERPARTY_AGENT_ADDRESS[-5:]} of payment.",
        )

    def test_handle_inform_with_data(self):
        """Test the _handle_inform method of the fipa handler where info has data."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:4],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.INFORM,
            info={"data_name": "data"},
        )

        # before
        for end_state_numbers in list(
            self.fipa_dialogues.dialogue_stats.self_initiated.values()
        ) + list(self.fipa_dialogues.dialogue_stats.other_initiated.values()):
            assert end_state_numbers == 0

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received INFORM from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )
        mock_logger.assert_any_call(
            logging.INFO, "received the following data={'data_name': 'data'}"
        )

        for (
            end_state_numbers
        ) in self.fipa_dialogues.dialogue_stats.other_initiated.values():
            assert end_state_numbers == 0
        for (
            end_state,
            end_state_numbers,
        ) in self.fipa_dialogues.dialogue_stats.self_initiated.items():
            if end_state == FipaDialogue.EndState.SUCCESSFUL:
                assert end_state_numbers == 1
            else:
                assert end_state_numbers == 0

    def test_handle_inform_without_data(self):
        """Test the _handle_inform method of the fipa handler where info has NO data."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:4],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.INFORM,
            info={},
        )

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received INFORM from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"received no data from sender={COUNTERPARTY_AGENT_ADDRESS[-5:]}",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the fipa handler."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages[:2],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.ACCEPT,
        )

        # operation
        with patch.object(self.fipa_handler.context.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle fipa message of performative={incoming_message.performative} in dialogue={fipa_dialogue}.",
        )

    def test_teardown(self):
        """Test the teardown method of the fipa handler."""
        assert self.fipa_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestGenericOefSearchHandler(BaseSkillTestCase):
    """Test oef search handler of generic buyer."""

    path_to_skill = PACKAGE_ROOT
    is_agent_to_agent_messages = False

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.oef_search_handler = cast(
            GenericOefSearchHandler, cls._skill.skill_context.handlers.oef_search
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)
        cls.oef_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.list_of_messages = (
            DialogueMessage(
                OefSearchMessage.Performative.SEARCH_SERVICES, {"query": "some_query"}
            ),
        )

    def setup_method(self) -> None:
        """Setup."""
        super().setup_method()
        self._init_strategy_kwargs = self.strategy.__dict__.copy()

    def teardown_method(self) -> None:
        """Teardown"""
        super().teardown_method()
        self.strategy.__dict__.update(self._init_strategy_kwargs)

    def test_setup(self):
        """Test the setup method of the oef_search handler."""
        assert self.oef_search_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the oef_search handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid oef_search message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the oef_search handler."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues,
            messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=oef_dialogue,
            performative=OefSearchMessage.Performative.OEF_ERROR,
            oef_error_operation=OefSearchMessage.OefErrorOperation.SEARCH_SERVICES,
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received oef_search error message={incoming_message} in dialogue={oef_dialogue}.",
        )

    def test_handle_search_zero_agents(self):
        """Test the _handle_search method of the oef_search handler."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues,
            messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=oef_dialogue,
            performative=OefSearchMessage.Performative.SEARCH_RESULT,
            agents=tuple(),
            agents_info=OefSearchMessage.AgentsInfo({}),
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"found no agents in dialogue={oef_dialogue}, continue searching.",
        )

    def test_handle_search_i(self):
        """Test the _handle_search method of the oef_search handler where is_stop_searching_on_result is True."""
        # setup
        self.strategy._max_negotiations = 3
        self.strategy._is_stop_searching_on_result = True
        self.strategy._is_searching = True

        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues,
            messages=self.list_of_messages[:1],
        )
        agents = ("agnt1", "agnt2")
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=oef_dialogue,
            performative=OefSearchMessage.Performative.SEARCH_RESULT,
            agents=agents,
            agents_info=OefSearchMessage.AgentsInfo(
                {"agent_1": {"key_1": "value_1"}, "agent_2": {"key_2": "value_2"}}
            ),
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"found agents={list(agents)}, stopping search."
        )

        assert self.strategy.is_searching is False

        self.assert_quantity_in_outbox(len(agents))
        for agent in agents:
            has_attributes, error_str = self.message_has_attributes(
                actual_message=self.get_message_from_outbox(),
                message_type=FipaMessage,
                performative=FipaMessage.Performative.CFP,
                to=agent,
                sender=self.skill.skill_context.agent_address,
                target=0,
                query=self.strategy.get_service_query(),
            )
            assert has_attributes, error_str
            mock_logger.assert_any_call(logging.INFO, f"sending CFP to agent={agent}")

    def test_handle_search_ii(self):
        """Test the _handle_search method of the oef_search handler where is_stop_searching_on_result is False."""
        # setup
        self.strategy._max_negotiations = 3
        self.strategy._is_stop_searching_on_result = False
        self.strategy._is_searching = True

        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues,
            messages=self.list_of_messages[:1],
        )
        agents = ("agnt1", "agnt2")
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=oef_dialogue,
            performative=OefSearchMessage.Performative.SEARCH_RESULT,
            agents=agents,
            agents_info=OefSearchMessage.AgentsInfo(
                {"agent_1": {"key_1": "value_1"}, "agent_2": {"key_2": "value_2"}}
            ),
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, f"found agents={list(agents)}.")

        assert self.strategy.is_searching is True

        self.assert_quantity_in_outbox(len(agents))
        for agent in agents:
            has_attributes, error_str = self.message_has_attributes(
                actual_message=self.get_message_from_outbox(),
                message_type=FipaMessage,
                performative=FipaMessage.Performative.CFP,
                to=agent,
                sender=self.skill.skill_context.agent_address,
                target=0,
                query=self.strategy.get_service_query(),
            )
            assert has_attributes, error_str
            mock_logger.assert_any_call(logging.INFO, f"sending CFP to agent={agent}")

    def test_handle_search_more_than_max_negotiation(self):
        """Test the _handle_search method of the oef_search handler where number of agents is more than max_negotiation."""
        # setup
        self.strategy._max_negotiations = 1
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues,
            messages=self.list_of_messages[:1],
        )
        agents = ("agnt1", "agnt2")
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=oef_dialogue,
            performative=OefSearchMessage.Performative.SEARCH_RESULT,
            agents=agents,
            agents_info=OefSearchMessage.AgentsInfo(
                {"agent_1": {"key_1": "value_1"}, "agent_2": {"key_2": "value_2"}}
            ),
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"found agents={list(agents)}, stopping search."
        )

        assert not self.strategy.is_searching

        self.assert_quantity_in_outbox(self.strategy._max_negotiations)
        for idx in range(0, self.strategy._max_negotiations):
            has_attributes, error_str = self.message_has_attributes(
                actual_message=self.get_message_from_outbox(),
                message_type=FipaMessage,
                performative=FipaMessage.Performative.CFP,
                to=agents[idx],
                sender=self.skill.skill_context.agent_address,
                target=0,
                query=self.strategy.get_service_query(),
            )
            assert has_attributes, error_str
            mock_logger.assert_any_call(
                logging.INFO, f"sending CFP to agent={agents[idx]}"
            )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the oef_search handler."""
        # setup
        invalid_performative = OefSearchMessage.Performative.UNREGISTER_SERVICE
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            service_description="some_service_description",
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle oef_search message of performative={invalid_performative} in dialogue={self.oef_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the oef_search handler."""
        assert self.oef_search_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestGenericSigningHandler(BaseSkillTestCase):
    """Test signing handler of generic buyer."""

    path_to_skill = PACKAGE_ROOT
    is_agent_to_agent_messages = False

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.signing_handler = cast(
            GenericSigningHandler, cls._skill.skill_context.handlers.signing
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.terms = Terms(
            "some_ledger_id",
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.list_of_fipa_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )
        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {
                    "terms": cls.terms,
                    "raw_transaction": SigningMessage.RawTransaction(
                        "some_ledger_id", {"some_key": "some_value"}
                    ),
                },
            ),
        )
        cls.list_of_ledger_api_messages = (
            DialogueMessage(LedgerApiMessage.Performative.GET_RAW_TRANSACTION, {}),
            DialogueMessage(LedgerApiMessage.Performative.RAW_TRANSACTION, {}),
            DialogueMessage(LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION, {}),
            DialogueMessage(LedgerApiMessage.Performative.TRANSACTION_DIGEST, {}),
        )

    def test_setup(self):
        """Test the setup method of the signing handler."""
        assert self.signing_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the signing handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=SigningMessage.Performative.ERROR,
            error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.signing_handler.context.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid signing message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_signed_transaction_last_ledger_api_message_is_none(
        self,
    ):
        """Test the _handle_signed_transaction method of the signing handler."""
        # setup
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:1],
            ),
        )
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:2],
            ),
        )
        signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue
        signing_dialogue.associated_ledger_api_dialogue._incoming_messages = []
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=signing_dialogue,
            performative=SigningMessage.Performative.SIGNED_TRANSACTION,
            signed_transaction=SigningMessage.SignedTransaction(
                "some_ledger_id", {"some_key": "some_value"}
            ),
        )

        # operation
        with pytest.raises(
            ValueError, match="Could not retrieve last message in ledger api dialogue"
        ):
            with patch.object(
                self.signing_handler.context.logger, "log"
            ) as mock_logger:
                self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, "transaction signing was successful.")

    def test_handle_signed_transaction_last_ledger_api_message_is_not_none(
        self,
    ):
        """Test the _handle_signed_transaction method of the signing handler where the last ledger_api message is not None."""
        # setup
        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:1],
                counterparty=signing_counterparty,
            ),
        )
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[0:2],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue
        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    "some_ledger_id", {"some_key": "some_value"}
                ),
            ),
        )
        # operation
        with patch.object(self.signing_handler.context.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, "transaction signing was successful.")

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            signed_transaction=incoming_message.signed_transaction,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "sending transaction to ledger.")

    def test_handle_error(self):
        """Test the _handle_error method of the signing handler."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_fipa_messages[:4],
                counterparty=COUNTERPARTY_AGENT_ADDRESS,
                is_agent_to_agent_messages=True,
            ),
        )

        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:1],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue

        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:1],
                counterparty=signing_counterparty,
            ),
        )
        signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.ERROR,
                error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
            ),
        )

        # operation
        with patch.object(
            self.signing_handler.context.behaviours.transaction, "failed_processing"
        ):
            with patch.object(
                self.signing_handler.context.logger, "log"
            ) as mock_logger:
                self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction signing was not successful. Error_code={incoming_message.error_code} in dialogue={signing_dialogue}",
        )

        behaviour = cast(
            GenericTransactionBehaviour, self.skill.skill_context.behaviours.transaction
        )

        # finish_processing
        assert behaviour.processing_time == 0.0
        assert behaviour.processing is None

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the signing handler."""
        # setup
        invalid_performative = SigningMessage.Performative.SIGN_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            terms=self.terms,
            raw_transaction=SigningMessage.RawTransaction(
                "some_ledger_id", {"some_key": "some_value"}
            ),
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.signing_handler.context.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle signing message of performative={invalid_performative} in dialogue={self.signing_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the signing handler."""
        assert self.signing_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestGenericLedgerApiHandler(BaseSkillTestCase):
    """Test ledger_api handler of generic buyer."""

    path_to_skill = PACKAGE_ROOT
    is_agent_to_agent_messages = False

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.ledger_api_handler = cast(
            GenericLedgerApiHandler, cls._skill.skill_context.handlers.ledger_api
        )
        cls.transaction_behaviour = cast(
            GenericTransactionBehaviour, cls._skill.skill_context.behaviours.transaction
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)
        cls.logger = cls._skill.skill_context.logger

        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )

        cls.terms = Terms(
            "some_ledger_id",
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.list_of_fipa_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )
        cls.raw_transaction = RawTransaction(
            "some_ledger_id", {"some_key": "some_value"}
        )
        cls.signed_transaction = SignedTransaction(
            "some_ledger_id", {"some_key": "some_value"}
        )
        cls.transaction_digest = TransactionDigest("some_ledger_id", "some_body")
        cls.transaction_receipt = TransactionReceipt(
            "some_ledger_id",
            {"receipt_key": "receipt_value"},
            {"transaction_key": "transaction_value"},
        )
        cls.list_of_ledger_api_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_RAW_TRANSACTION, {"terms": cls.terms}
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.RAW_TRANSACTION,
                {"raw_transaction": cls.raw_transaction},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                {"signed_transaction": cls.signed_transaction},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                {"transaction_digest": cls.transaction_digest},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                {"transaction_digest": cls.transaction_digest},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                {"transaction_receipt": cls.transaction_receipt},
            ),
        )

    def test_setup(self):
        """Test the setup method of the ledger_api handler."""
        assert self.ledger_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the ledger_api handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id="some_ledger_id",
            address="some_address",
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid ledger_api message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_balance_positive_balance(self):
        """Test the _handle_balance method of the ledger_api handler where balance is positive."""
        # setup
        balance = 10
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=(
                    DialogueMessage(
                        LedgerApiMessage.Performative.GET_BALANCE,
                        {"ledger_id": "some_ledger_id", "address": "some_address"},
                    ),
                ),
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.BALANCE,
                ledger_id="some-Ledger_id",
                balance=balance,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"starting balance on {self.strategy.ledger_id} ledger={incoming_message.balance}.",
        )
        assert self.strategy.balance == balance
        assert self.strategy.is_searching

    def test_handle_balance_zero_balance(self):
        """Test the _handle_balance method of the ledger_api handler where balance is zero."""
        # setup
        balance = 0
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=(
                    DialogueMessage(
                        LedgerApiMessage.Performative.GET_BALANCE,
                        {"ledger_id": "some_ledger_id", "address": "some_address"},
                    ),
                ),
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.BALANCE,
                ledger_id="some-Ledger_id",
                balance=balance,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"you have no starting balance on {self.strategy.ledger_id} ledger! Stopping skill {self.strategy.context.skill_id}.",
        )
        assert not self.skill.skill_context.is_active

    def test_handle_raw_transaction(self):
        """Test the _handle_raw_transaction method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:1],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_fipa_messages[:4],
                is_agent_to_agent_messages=True,
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        fipa_dialogue.terms = self.terms
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.RAW_TRANSACTION,
                raw_transaction=self.raw_transaction,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"received raw transaction={incoming_message}"
        )

        message_quantity = self.get_quantity_in_decision_maker_inbox()
        assert (
            message_quantity == 1
        ), f"Invalid number of messages in decision maker queue. Expected {1}. Found {message_quantity}."
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_decision_maker_inbox(),
            message_type=SigningMessage,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            terms=self.terms,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO,
            "proposing the transaction to the decision maker. Waiting for confirmation ...",
        )

    def test_handle_transaction_digest(self):
        """Test the _handle_transaction_digest method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[2:3],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                transaction_digest=self.transaction_digest,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully submitted. Transaction digest={incoming_message.transaction_digest}",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
            to=incoming_message.sender,
            sender=str(self.skill.skill_context.skill_id),
            transaction_digest=self.transaction_digest,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO,
            "checking transaction is settled.",
        )

    def test_handle_transaction_receipt_i(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[4:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_fipa_messages[:4],
                is_agent_to_agent_messages=True,
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        fipa_dialogue.terms = self.terms
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(
            self.ledger_api_handler.context.behaviours.transaction, "finish_processing"
        ):
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                with patch.object(self.logger, "log") as mock_logger:
                    self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction confirmed, informing counterparty={fipa_dialogue.dialogue_label.dialogue_opponent_addr[-5:]} of transaction digest.",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=FipaMessage,
            performative=FipaMessage.Performative.INFORM,
            to=COUNTERPARTY_AGENT_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            info={"transaction_digest": self.transaction_digest.body},
        )
        assert has_attributes, error_str

    def test_handle_transaction_receipt_ii(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where fipa dialogue's last_incoming_message is None."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[4:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_fipa_messages[:4],
                is_agent_to_agent_messages=True,
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue

        fipa_dialogue._incoming_messages = []

        fipa_dialogue.terms = self.terms
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(
            self.ledger_api_handler.context.behaviours.transaction, "finish_processing"
        ):
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                with patch.object(self.logger, "log"):
                    with pytest.raises(
                        ValueError, match="Could not retrieve last fipa message"
                    ):
                        self.ledger_api_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(0)

    def test_handle_transaction_receipt_iii(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where tx is NOT settled."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[4:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_fipa_messages[:4],
                is_agent_to_agent_messages=True,
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        fipa_dialogue.terms = self.terms
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )
        # operation
        with patch.object(
            self.ledger_api_handler.context.behaviours.transaction, "failed_processing"
        ):
            with patch.object(LedgerApis, "is_transaction_settled", return_value=False):
                with patch.object(self.logger, "log") as mock_logger:
                    self.ledger_api_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(0)
        assert self.transaction_behaviour.processing is None
        assert self.transaction_behaviour.processing_time == 0.0

        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction_receipt={self.transaction_receipt} not settled or not valid, aborting",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = self.prepare_skill_dialogue(
            dialogues=self.ledger_api_dialogues,
            messages=self.list_of_ledger_api_messages[:1],
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.ERROR,
                code=1,
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = "mock"
        # operation
        with patch.object(
            self.ledger_api_handler.context.behaviours.transaction, "failed_processing"
        ):
            with patch.object(self.logger, "log") as mock_logger:
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received ledger_api error message={incoming_message} in dialogue={ledger_api_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the ledger_api handler."""
        # setup
        invalid_performative = LedgerApiMessage.Performative.GET_BALANCE
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            ledger_id="some_ledger_id",
            address="some_address",
            to=str(self.skill.public_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle ledger_api message of performative={invalid_performative} in dialogue={self.ledger_api_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the ledger_api handler."""
        assert self.ledger_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the CLI plugin mechanism."""

import inspect
from importlib.metadata import Distribution, EntryPoint
from pathlib import Path

import click
import pytest

# Create a few CLI commands for testing
from aea.cli.plugin import with_plugins
from aea.test_tools.click_testing import CliRunner

from tests.conftest import ROOT_DIR

# We need to compute the right dotted path w.r.t. the current module location.
# Instead of hard-coding it, we compute it on-the-fly.
_PATH_TO_THIS_MODULE = (
    Path(inspect.getfile(inspect.currentframe()))  # type: ignore
    .absolute()
    .relative_to(Path(ROOT_DIR).resolve())
)
_DOTTED_PATH = ".".join(_PATH_TO_THIS_MODULE.with_suffix("").parts)


@pytest.fixture(scope="function")
def runner(request):
    """Get a click.CliRunner instance."""
    return CliRunner()


@click.command()
@click.argument("arg")
def cmd1(arg):
    """Test command 1"""
    click.echo("passed")


@click.command()
@click.argument("arg")
def cmd2(arg):
    """Test command 2"""
    click.echo("passed")


class DistStub(Distribution):
    """
    Manually register plugins in an entry point and put broken plugins in a different entry point.

    The `DistStub()` class gets around an exception that is raised when
    `entry_point.load()` is called.  By default `load()` has `requires=True`
    which calls `dist.requires()` and the `click.group()` decorator
    doesn't allow us to change this.  Because we are manually registering these
    plugins the `dist` attribute is `None` so we can just create a stub that
    always returns an empty list since we don't have any requirements.  A full
    `Distribution()` instance is not needed because there isn't
    a package installed anywhere.
    """

    def requires(self, *args):
        """Implement a dummy 'requires' function."""
        return []


# Note: This test setup is using internal pkg_resources structures that don't have
# direct equivalents in importlib.metadata, so we'll need to mock the entry_points
# function instead for testing
def _mock_entry_points(group: str):
    """Mock entry_points for testing."""
    test_entries = {
        "_test_click_plugins.test_plugins": [
            EntryPoint(name="cmd1", value=f"{_DOTTED_PATH}:cmd1", group=group),
            EntryPoint(name="cmd2", value=f"{_DOTTED_PATH}:cmd2", group=group),
        ],
        "_test_click_plugins.broken_plugins": [
            EntryPoint(name="before", value="tests.broken_plugins:before", group=group),
            EntryPoint(name="after", value="tests.broken_plugins:after", group=group),
            EntryPoint(
                name="do_not_exist",
                value="tests.broken_plugins:do_not_exist",
                group=group,
            ),
        ],
    }
    return test_entries.get(group, [])


# Main CLI groups - one with good plugins attached and the other broken
@with_plugins(_mock_entry_points("_test_click_plugins.test_plugins"))
@click.group()
def good_cli():
    """Good CLI group."""
    pass


@with_plugins(_mock_entry_points("_test_click_plugins.broken_plugins"))
@click.group()
def broken_cli():
    """Broken CLI group."""
    pass


def test_registered():
    """
    Make sure the plugins are properly registered.

    If this test fails it means that some of the for loops in other tests may not be executing.
    """
    assert len(_mock_entry_points("_test_click_plugins.test_plugins")) > 1
    assert len(_mock_entry_points("_test_click_plugins.broken_plugins")) > 1


def test_register_and_run(runner):
    """Test that registration and run of the command work correctly."""

    result = runner.invoke(good_cli, ["--help"])
    assert result.exit_code == 0

    for ep in _mock_entry_points("_test_click_plugins.test_plugins"):
        cmd_result = runner.invoke(good_cli, [ep.name, "something"])
        assert cmd_result.exit_code == 0
        assert cmd_result.output.strip() == "passed"


def test_broken_register_and_run(runner):
    """Test that the broken plugin doesn't get registered as expected."""
    result = runner.invoke(broken_cli, ["--help"])
    assert result.exit_code == 0

    for ep in _mock_entry_points("_test_click_plugins.broken_plugins"):
        cmd_result = runner.invoke(broken_cli, [ep.name])
        assert cmd_result.exit_code == 1
        assert "Traceback" in cmd_result.output


def test_group_chain(runner):
    """Test the plugin with nested CLI command group levels."""

    # Attach a sub-group to a CLI and get execute it without arguments to make
    # sure both the sub-group and all the parent group's commands are present
    @good_cli.group()
    def sub_cli():
        """Sub CLI."""
        pass

    result = runner.invoke(good_cli, ["--help"])
    assert result.exit_code == 0
    assert sub_cli.name in result.output
    for ep in _mock_entry_points("_test_click_plugins.test_plugins"):
        assert ep.name in result.output

    # Same as above but the sub-group has plugins
    @with_plugins(plugins=_mock_entry_points("_test_click_plugins.test_plugins"))
    @good_cli.group(name="sub-cli-plugins")
    def sub_cli_plugins():
        """Sub CLI with plugins."""
        pass

    result = runner.invoke(good_cli, ["sub-cli-plugins", "--help"])
    assert result.exit_code == 0
    for ep in _mock_entry_points("_test_click_plugins.test_plugins"):
        assert ep.name in result.output

    print(result.output)

    # Execute one of the sub-group's commands
    result = runner.invoke(good_cli, ["sub-cli-plugins", "cmd1", "something"])
    assert result.exit_code == 0
    assert result.output.strip() == "passed"


def test_exception():
    """Test the 'with_plugins' decorator when it gets used on a non-click.Group object."""

    # Decorating something that isn't a click.Group() should fail
    with pytest.raises(TypeError):

        @with_plugins([])
        @click.command()
        def cli():
            """Whatever"""


def test_broken_register_and_run_with_help(runner):
    """Test the broken registration of the plugin when the command is run with the '--help' flag."""
    result = runner.invoke(broken_cli, ["--help"])
    assert result.exit_code == 0

    for ep in _mock_entry_points("_test_click_plugins.broken_plugins"):
        cmd_result = runner.invoke(broken_cli, [ep.name, "--help"])
        assert cmd_result.exit_code == 1
        assert "Traceback" in cmd_result.output


def test_broken_register_and_run_with_args(runner):
    """Test the broken registration of the plugin when the command is run with the '--help' flag."""
    result = runner.invoke(broken_cli, ["--help"])
    assert result.exit_code == 0

    for ep in _mock_entry_points("_test_click_plugins.broken_plugins"):
        cmd_result = runner.invoke(broken_cli, [ep.name, "-a", "b"])
        assert cmd_result.exit_code == 1
        assert "Traceback" in cmd_result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the generic buyer skill."""

# pylint: skip-file

from pathlib import Path
from typing import cast

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.transaction.base import Terms
from aea.protocols.dialogue.base import DialogueLabel
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.generic_buyer.dialogues import (
    DefaultDialogue,
    DefaultDialogues,
    FipaDialogue,
    FipaDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    OefSearchDialogue,
    OefSearchDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.open_aea.protocols.signing.message import SigningMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

PACKAGE_ROOT = Path(__file__).parent.parent


class TestDialogues(BaseSkillTestCase):
    """Test dialogue classes of generic buyer."""

    path_to_skill = PACKAGE_ROOT

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )

    def test_default_dialogues(self):
        """Test the DefaultDialogues class."""
        _, dialogue = self.default_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=DefaultMessage.Performative.BYTES,
            content=b"some_content",
        )
        assert dialogue.role == DefaultDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_fipa_dialogue(self):
        """Test the FipaDialogue class."""
        fipa_dialogue = FipaDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=DefaultDialogue.Role.AGENT,
        )

        # terms
        with pytest.raises(AEAEnforceError, match="Terms not set!"):
            assert fipa_dialogue.terms
        terms = Terms(
            "some_ledger_id",
            self.skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        fipa_dialogue.terms = terms
        with pytest.raises(AEAEnforceError, match="Terms already set!"):
            fipa_dialogue.terms = terms
        assert fipa_dialogue.terms == terms

    def test_fipa_dialogues(self):
        """Test the FipaDialogues class."""
        _, dialogue = self.fipa_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=FipaMessage.Performative.CFP,
            query="some_query",
        )
        assert dialogue.role == FipaDialogue.Role.BUYER
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_ledger_api_dialogue(self):
        """Test the LedgerApiDialogue class."""
        ledger_api_dialogue = LedgerApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=LedgerApiDialogue.Role.AGENT,
        )

        # associated_fipa_dialogue
        with pytest.raises(AEAEnforceError, match="FipaDialogue not set!"):
            assert ledger_api_dialogue.associated_fipa_dialogue
        fipa_dialogue = FipaDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=FipaDialogue.Role.BUYER,
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        with pytest.raises(AEAEnforceError, match="FipaDialogue already set!"):
            ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        assert ledger_api_dialogue.associated_fipa_dialogue == fipa_dialogue

    def test_ledger_api_dialogues(self):
        """Test the LedgerApiDialogues class."""
        _, dialogue = self.ledger_api_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id="some_ledger_id",
            address="some_address",
        )
        assert dialogue.role == LedgerApiDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.public_id)

    def test_oef_search_dialogues(self):
        """Test the OefSearchDialogues class."""
        _, dialogue = self.oef_search_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query="some_query",
        )
        assert dialogue.role == OefSearchDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.public_id)

    def test_signing_dialogue(self):
        """Test the SigningDialogue class."""
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=SigningDialogue.Role.SKILL,
        )

        # associated_ledger_api_dialogue
        with pytest.raises(AEAEnforceError, match="LedgerApiDialogue not set!"):
            assert signing_dialogue.associated_ledger_api_dialogue
        ledger_api_dialogue = LedgerApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=LedgerApiDialogue.Role.AGENT,
        )
        signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue
        with pytest.raises(AEAEnforceError, match="LedgerApiDialogue already set!"):
            signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue
        assert signing_dialogue.associated_ledger_api_dialogue == ledger_api_dialogue

    def test_signing_dialogues(self):
        """Test the SigningDialogues class."""
        _, dialogue = self.signing_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            terms="some_terms",
            raw_transaction="some_raw_transaction",
        )
        assert dialogue.role == SigningDialogue.Role.SKILL
        assert dialogue.self_address == str(self.skill.public_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains test case classes based on pytest for AEA contract testing."""

import time
from abc import ABC, ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, cast

from aea.common import JSONLike
from aea.configurations.loader import (
    ComponentType,
    ContractConfig,
    load_component_configuration,
)
from aea.contracts.base import Contract, contract_registry
from aea.crypto.base import Crypto, FaucetApi, LedgerApi
from aea.crypto.registries import (
    crypto_registry,
    faucet_apis_registry,
    ledger_apis_registry,
)


class _MetaBaseContractTestCase(ABCMeta):
    """A metaclass that validates BaseContractTestCase's class."""

    def __new__(  # type: ignore # pylint: disable=bad-mcs-classmethod-argument # type: ignore
        mcs, name: str, bases: Tuple, namespace: Dict, **kwargs: Any
    ) -> Type:
        """Initialize the class."""
        new_cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        if ABC in bases:
            # abstract class, return
            return new_cls
        if not issubclass(new_cls, BaseContractTestCase):
            # the check only applies to BaseContractTestCase subclasses
            return new_cls

        mcs._check_consistency(cast(Type[BaseContractTestCase], new_cls))
        return new_cls

    @classmethod
    def _check_consistency(  # pylint: disable=bad-mcs-classmethod-argument
        mcs, base_contract_test_case_cls: Type["BaseContractTestCase"]
    ) -> None:
        """Check consistency of class."""
        mcs._check_required_class_attributes(base_contract_test_case_cls)

    @classmethod
    def _check_required_class_attributes(  # pylint: disable=bad-mcs-classmethod-argument
        mcs, base_contract_test_case_cls: Type["BaseContractTestCase"]
    ) -> None:
        """Check that required class attributes are set."""
        if not hasattr(base_contract_test_case_cls, "path_to_contract"):
            raise ValueError(
                f"'path_to_contract' not set on {base_contract_test_case_cls}"
            )
        if not hasattr(base_contract_test_case_cls, "ledger_identifier"):
            raise ValueError(
                f"'ledger_identifier' not set on {base_contract_test_case_cls}"
            )


class BaseContractTestCase(ABC, metaclass=_MetaBaseContractTestCase):
    """A class to test a contract."""

    path_to_contract: Path
    ledger_identifier: str
    fund_from_faucet: bool = False
    _deployment_gas: int = 5000000
    _contract: Contract

    ledger_api: LedgerApi
    deployer_crypto: Crypto
    item_owner_crypto: Crypto
    faucet_api: FaucetApi

    deployment_tx_receipt: JSONLike
    contract_address: str

    @property
    def contract(self) -> Contract:
        """Get the contract."""
        try:
            value = self._contract
        except AttributeError:
            raise ValueError("Ensure the contract is set during setup.")
        return value

    @classmethod
    def setup_class(cls) -> None:
        """Set up the contract test case class."""
        # register contract
        configuration = cast(
            ContractConfig,
            load_component_configuration(ComponentType.CONTRACT, cls.path_to_contract),
        )
        configuration._directory = (  # pylint: disable=protected-access
            cls.path_to_contract
        )
        if str(configuration.public_id) not in contract_registry.specs:
            # load contract into sys modules
            Contract.from_config(configuration)  # pragma: nocover
        cls._contract = contract_registry.make(str(configuration.public_id))

    @classmethod
    def setup_method(cls, **kwargs: Any) -> None:
        """Set up the contract test case."""
        _ledger_config: Dict[str, str] = kwargs.pop("ledger_config", {})
        _deployer_private_key_path: Optional[str] = kwargs.pop(
            "deployer_private_key_path", None
        )
        _item_owner_private_key_path: Optional[str] = kwargs.pop(
            "item_owner_private_key_path", None
        )
        cls.ledger_api = ledger_apis_registry.make(
            cls.ledger_identifier, **_ledger_config
        )
        cls.deployer_crypto = crypto_registry.make(
            cls.ledger_identifier, private_key_path=_deployer_private_key_path
        )
        cls.item_owner_crypto = crypto_registry.make(
            cls.ledger_identifier, private_key_path=_item_owner_private_key_path
        )
        cls.faucet_api = faucet_apis_registry.make(cls.ledger_identifier)

        # Fund from faucet
        if cls.fund_from_faucet:
            # Refill deployer account from faucet
            cls.refill_from_faucet(
                cls.ledger_api, cls.faucet_api, cls.deployer_crypto.address
            )

            # Refill item owner account from faucet
            cls.refill_from_faucet(
                cls.ledger_api, cls.faucet_api, cls.item_owner_crypto.address
            )

        # deploy contract
        cls.deployment_tx_receipt = cls._deploy_contract(
            cls._contract, cls.ledger_api, cls.deployer_crypto, gas=cls._deployment_gas
        )
        cls.contract_address = cls.finish_contract_deployment()

    @classmethod
    @abstractmethod
    def finish_contract_deployment(cls) -> str:
        """
        Finish deploying contract.

        :return: contract address
        """

    @staticmethod
    def refill_from_faucet(
        ledger_api: LedgerApi, faucet_api: FaucetApi, address: str
    ) -> None:
        """Refill from faucet."""
        start_balance = ledger_api.get_balance(address)

        faucet_api.get_wealth(address)

        balance = ledger_api.get_balance(address)
        if balance == start_balance:
            raise ValueError("Balance not increased!")  # pragma: nocover

    @staticmethod
    def sign_send_confirm_receipt_multisig_transaction(
        tx: JSONLike,
        ledger_api: LedgerApi,
        cryptos: List[Crypto],
        sleep_time: float = 2.0,
    ) -> JSONLike:
        """
        Sign, send and confirm settlement of a transaction with multiple signatures.

        :param tx: the transaction
        :param ledger_api: the ledger api
        :param cryptos: Cryptos to sign transaction with
        :param sleep_time: the time to sleep between transaction submission and receipt request
        :return: The transaction receipt
        """
        for crypto in cryptos:
            tx = crypto.sign_transaction(tx)
        tx_digest = ledger_api.send_signed_transaction(tx)

        if tx_digest is None:
            raise ValueError("Transaction digest not found!")  # pragma: nocover

        tx_receipt = ledger_api.get_transaction_receipt(tx_digest)

        not_settled = True
        elapsed_time = 0
        while not_settled and elapsed_time < 20:
            elapsed_time += 1
            time.sleep(sleep_time)
            tx_receipt = ledger_api.get_transaction_receipt(tx_digest)
            if tx_receipt is None:
                continue
            not_settled = not ledger_api.is_transaction_settled(tx_receipt)

        if tx_receipt is None:
            raise ValueError("Transaction receipt not found!")  # pragma: nocover

        if not_settled:
            raise ValueError(  # pragma: nocover
                f"Transaction receipt not valid!\n{tx_receipt['raw_log']}"
            )

        return tx_receipt

    @classmethod  # noqa
    def sign_send_confirm_receipt_transaction(
        cls,
        tx: JSONLike,
        ledger_api: LedgerApi,
        crypto: Crypto,
        sleep_time: float = 2.0,
    ) -> JSONLike:
        """
        Sign, send and confirm settlement of a transaction with multiple signatures.

        :param tx: the transaction
        :param ledger_api: the ledger api
        :param crypto: Crypto to sign transaction with
        :param sleep_time: the time to sleep between transaction submission and receipt request
        :return: The transaction receipt
        """

        # BACKWARDS COMPATIBILITY: This method supports only 1 signer and is kept for backwards compatibility.
        # new method sign_send_confirm_receipt_multisig_transaction should be used always instead of this one.
        return cls.sign_send_confirm_receipt_multisig_transaction(
            tx, ledger_api, [crypto], sleep_time
        )

    @classmethod
    def _deploy_contract(
        cls,
        contract: Contract,
        ledger_api: LedgerApi,
        deployer_crypto: Crypto,
        gas: int,
    ) -> JSONLike:
        """
        Deploy contract on network.

        :param contract: the contract
        :param ledger_api: the ledger api
        :param deployer_crypto: the contract deployer crypto
        :param gas: the gas amount

        :return: the transaction receipt for initial transaction deployment
        """
        tx = contract.get_deploy_transaction(
            ledger_api=ledger_api,
            deployer_address=deployer_crypto.address,
            gas=gas,
        )

        if tx is None:
            raise ValueError("Deploy transaction not found!")  # pragma: nocover

        tx_receipt = cls.sign_send_confirm_receipt_multisig_transaction(
            tx, ledger_api, [deployer_crypto]
        )

        return tx_receipt



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for tac protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.fetchai.protocols.tac.custom_types import ErrorCode
from packages.fetchai.protocols.tac.message import TacMessage


class TestMessageTac(BaseProtocolMessagesTestCase):
    """Test for the 'tac' protocol message."""

    MESSAGE_CLASS = TacMessage

    def build_messages(self) -> List[TacMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            TacMessage(
                performative=TacMessage.Performative.REGISTER,
                agent_name="some str",
            ),
            TacMessage(
                performative=TacMessage.Performative.UNREGISTER,
            ),
            TacMessage(
                performative=TacMessage.Performative.TRANSACTION,
                transaction_id="some str",
                ledger_id="some str",
                sender_address="some str",
                counterparty_address="some str",
                amount_by_currency_id={"some str": 12},
                fee_by_currency_id={"some str": 12},
                quantities_by_good_id={"some str": 12},
                nonce="some str",
                sender_signature="some str",
                counterparty_signature="some str",
            ),
            TacMessage(
                performative=TacMessage.Performative.CANCELLED,
            ),
            TacMessage(
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id={"some str": 12},
                exchange_params_by_currency_id={"some str": 1.0},
                quantities_by_good_id={"some str": 12},
                utility_params_by_good_id={"some str": 1.0},
                fee_by_currency_id={"some str": 12},
                agent_addr_to_name={"some str": "some str"},
                currency_id_to_name={"some str": "some str"},
                good_id_to_name={"some str": "some str"},
                version_id="some str",
                info={"some str": "some str"},
            ),
            TacMessage(
                performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
                transaction_id="some str",
                amount_by_currency_id={"some str": 12},
                quantities_by_good_id={"some str": 12},
            ),
            TacMessage(
                performative=TacMessage.Performative.TAC_ERROR,
                error_code=ErrorCode.TRANSACTION_NOT_MATCHING,
                info={"some str": "some str"},
            ),
        ]

    def build_inconsistent(self) -> List[TacMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            TacMessage(
                performative=TacMessage.Performative.REGISTER,
                # skip content: agent_name
            ),
            TacMessage(
                performative=TacMessage.Performative.TRANSACTION,
                # skip content: transaction_id
                ledger_id="some str",
                sender_address="some str",
                counterparty_address="some str",
                amount_by_currency_id={"some str": 12},
                fee_by_currency_id={"some str": 12},
                quantities_by_good_id={"some str": 12},
                nonce="some str",
                sender_signature="some str",
                counterparty_signature="some str",
            ),
            TacMessage(
                performative=TacMessage.Performative.GAME_DATA,
                # skip content: amount_by_currency_id
                exchange_params_by_currency_id={"some str": 1.4},
                quantities_by_good_id={"some str": 12},
                utility_params_by_good_id={"some str": 1.4},
                fee_by_currency_id={"some str": 12},
                agent_addr_to_name={"some str": "some str"},
                currency_id_to_name={"some str": "some str"},
                good_id_to_name={"some str": "some str"},
                version_id="some str",
                info={"some str": "some str"},
            ),
            TacMessage(
                performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
                # skip content: transaction_id
                amount_by_currency_id={"some str": 12},
                quantities_by_good_id={"some str": 12},
            ),
            TacMessage(
                performative=TacMessage.Performative.TAC_ERROR,
                # skip content: error_code
                info={"some str": "some str"},
            ),
        ]
